:- [utils/common].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Predicates provided from common.pl
%%
%% parsePDDL(+DomainFile, +ProblemFile)
%%    Parses the domain and problem PDDL files and initializes a set
%%    of global variables you can access to get information about the
%%    domain and problem as follows.
%%
%% get_goal(-Goal)
%%    Goal is the goal description of the parsed problem as a list
%%    of literals
%%
%% get_init(-InitialState)
%%    InitialState is the description of the initial state of the parsed
%%    problem as a list of literals
%%
%% satisfies_goal(+State)
%%    Checks if the variable State (that contains a list of literals)
%%    satisfies the goal.
%%
%% progress(+State, -Action, -NextState)
%%    Generates all the possible successor states from State. Action is
%%    an applicable ground action to State, and NextState is the
%%    successor state when the effects of Action are applied to State.
%%
%% h(+State, -Value)
%%    Heuristic function that estimates the distance of State to the
%%    a state that satisfies the goal, by counting the sub-goals that
%%    are not satisfied in State.
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% dfsplanner(+DomainFile, +ProblemFile)
%%    Parses the domain and problem PDDL files
%%    Prints information about the objects, initial state, goal, and
%%    available actions of the parsed domain and problem files
%%    Searches for a solution using the predicate search/1
%%    If a solution is found, print the solution and statistics on screen
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
dfsplanner(DomainFile, ProblemFile):-
        % Parse the PDDL files and put the necessary information
        % into global variables
        parsePDDL(DomainFile, ProblemFile),
        % Print useful information from the parsed files
        show_information,
        % Reset the global variables that hold statistic information
        % about the search
        reset_statistic,
        % Perform forward search to find a sequence of actions that
        % achieves the goal
        print('Planning...\n'), flush_output,
        dfssearch(Solution),
        % If we get up to here then a solution was found
        print('\nSolution found!\n\n'),
        show_statistic(Solution),
		print_statistic(Solution,'output.txt').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% dfssearch(-Solution)
%%
%% Initializes depth-first-search by calling dfs/3 with the appropriate
%% initial values. Binds the variable "Solution" to the solution that is
%% found by dfs.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
dfssearch(Solution):-
        get_init(InitialState),
        OpenStates = [[]-InitialState],
        VisitedStates = [],
        dfs(OpenStates, VisitedStates, Solution).

% Check if the first state in the list of open states satisfies the goal.
dfs(OpenStates, _VisitedStates, Solution):-
        % Get the first state in the list of open states.
        OpenStates = [GroundActions-State | _],
                %Debug information -- uncomment to print useful information
                %print('*Debug* check state if satisfies the goal:\n'), print(State), nl,
        % Check if the state description satisfies the goal.
        satisfies_goal(State),
        % If we get up to here, then State satisfies the goal and the solution
        % is the list of ground actions that get us to State
        reverse(GroundActions, GroundActionsReversed),
        Solution = GroundActionsReversed,
        % The search is over, so no more bactracking!
        !.

% Then check if the first state in the list of open states is already visited.
dfs(OpenStates, VisitedStates, Solution):-
        % Get the first state in the list of open states
        OpenStates = [_GroundActions-State | RestOfOpenStates],
                % Debug information -- uncomment to print useful information
                % print('*Debug* check state if it is already visited:\n'), print(State), nl,
        % Check if the state description is already in the list of visited states
        member(State, VisitedStates),
        % If we get up to here, then it's a state we've seen before.
        % No need to expand this node using the next rule, so no more bactracking!
        !,
                % Debug information -- uncomment to print useful information
                % print('*Debug* state is already visited:\n'), print(State), nl,
        % Continue search using the remaining list of open states.
        dfs(RestOfOpenStates, VisitedStates, Solution).

% Finally, expand the first state in the list of open states and recurse.
dfs(OpenStates, VisitedStates, Solution):-
        % Get the first state in the list of open states
        OpenStates = [GroundActions-State | RestOfOpenStates],
                % Debug information -- uncomment to print useful information
                % print('*Debug* expand state:\n'), print(State), nl,
                % Debug information -- uncomment to print useful information
                % print('*Debug* open states:\n'), print(OpenStates), nl,
        % Expand the state by inserting the next possible states in the list of open states
        findall([TempGroundAction | GroundActions ]-TempNextState,
	        progress(State, TempGroundAction, TempNextState),
                NextStates),
        % Print a dot for every node expanded.
        % print('.'), flush_output,
        % Add statistic information.
        stat_node,
        append(NextStates, RestOfOpenStates, NewOpenStates),
                 % Debug information -- uncomment to print useful information
                 % print('*Debug* expand state, new set of open states:\n'), print(NewOpenStates), nl,
        dfs(NewOpenStates, [State | VisitedStates], Solution).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% astarplanner(+DomainFile, +ProblemFile)
%%    Parses the domain and problem PDDL files
%%    Prints information about the objects, initial state, goal, and
%%    available actions of the parsed domain and problem files
%%    Searches for a solution using the predicate search/1
%%    If a solution is found, print the solution and statistics on screen
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
astarplanner(DomainFile, ProblemFile):-
        % Parse the PDDL files and put the necessary information
        % into global variables
        parsePDDL(DomainFile, ProblemFile),
        % Print useful information from the parsed files
        show_information,
        % Reset the global variables that hold statistic information
        % about the search
        reset_statistic,
        % Perform forward search to find a sequence of actions that
        % achieves the goal
        print('Planning...\n'), flush_output,
        astarsearch(Solution),
        % If we get up to here then a solution was found
        print('\nSolution found!\n\n'),
        show_statistic(Solution),
		print_statistic(Solution,'output.txt').

astarsearch(Solution):-
        get_init(InitialState),
        OpenStates = [0-[]-InitialState],
        VisitedStates = [],
        astar(OpenStates, VisitedStates, Solution).

% Check if the first state in the list of open states satisfies the goal.
astar(OpenStates, _VisitedStates, Solution):-
        % Get the first state in the list of open states.
        OpenStates = [_Value-GroundActions-State | _],
                %Debug information -- uncomment to print useful information
                %print('*Debug* check state if satisfies the goal:\n'), print(State), nl,
        % Check if the state description satisfies the goal.
        satisfies_goal(State),
        % If we get up to here, then State satisfies the goal and the solution
        % is the list of ground actions that get us to State
        reverse(GroundActions, GroundActionsReversed),
        Solution = GroundActionsReversed,
        % The search is over, so no more bactracking!
        !.

% Then check if the first state in the list of open states is already visited.
astar(OpenStates, VisitedStates, Solution):-
        % Get the first state in the list of open states
        OpenStates = [_Value-_GroundActions-State | RestOfOpenStates],
                % Debug information -- uncomment to print useful information
                % print('*Debug* check state if it is already visited:\n'), print(State), nl,
        % Check if the state description is already in the list of visited states
        member(State, VisitedStates),
        % If we get up to here, then it's a state we've seen before.
        % No need to expand this node using the next rule, so no more bactracking!
        !,
                % Debug information -- uncomment to print useful information
                % print('*Debug* state is already visited:\n'), print(State), nl,
        % Continue search using the remaining list of open states.
        astar(RestOfOpenStates, VisitedStates, Solution).

% Finally, expand the first state in the list of open states and recurse.
astar(OpenStates, VisitedStates, Solution):-
        % Get the first state in the list of open states
        OpenStates = [_Value-GroundActions-State | RestOfOpenStates],
                % Debug information -- uncomment to print useful information
                % print('*Debug* expand state:\n'), print(State), nl,
                % Debug information -- uncomment to print useful information
                % print('*Debug* open states:\n'), print(OpenStates), nl,
        % Depth of State is the size of the list of ground actions that get us there.
        length(GroundActions,Depth),
        % Expand the state by inserting the next possible states in the list of open states
        findall(TempValue-[TempGroundAction | GroundActions ]-TempNextState,
                best_step(State, Depth, TempGroundAction, TempNextState, TempValue),
                NextStates),
        % Print a dot for every node expanded.
        % print('.'), flush_output,
        % Add statistic information.
        stat_node,
        append(NextStates, RestOfOpenStates, NewOpenStates),
                 % Debug information -- uncomment to print useful information
                 % print('*Debug* expand state, new set of open states:\n'), print(NewOpenStates), nl,
        keysort(NewOpenStates, OrderedNewOpenStates),
        astar(OrderedNewOpenStates, [State | VisitedStates], Solution).



%best_step(State, Deep, ActionDef, NewState, EstimatedValue):-
best_step(S, D, ActionDef, NewS, E):-
        progress(S, ActionDef, NewS),
        h(NewS, V),
        E is D+V.





























































